/* eslint-disable class-methods-use-this,@typescript-eslint/no-unused-vars */
import { Sequelize } from 'sequelize'
import { IRepositoryOptions } from './IRepositoryOptions'
import { PageData, SearchCriteria } from '../../types/common'
import AuditLogRepository from './auditLogRepository'
import SequelizeRepository from './sequelizeRepository'

export abstract class RepositoryBase<
  TData,
  TId,
  TCreate,
  TUpdate,
  TCriteria extends SearchCriteria,
> {
  protected constructor(
    public readonly options: IRepositoryOptions,
    public readonly log: boolean = false,
  ) {}

  protected get currentUser(): any {
    return SequelizeRepository.getCurrentUser(this.options)
  }

  protected get currentTenant(): any {
    return SequelizeRepository.getCurrentTenant(this.options)
  }

  protected get transaction(): any {
    return SequelizeRepository.getTransaction(this.options)
  }

  protected get seq(): Sequelize {
    return SequelizeRepository.getSequelize(this.options)
  }

  protected get database(): any {
    return this.options.database
  }

  abstract create(data: TCreate): Promise<TData>

  abstract update(id: TId, data: TUpdate): Promise<TData>

  async destroy(id: TId): Promise<void> {
    return this.destroyAll([id])
  }

  abstract destroyAll(ids: TId[]): Promise<void>

  abstract findById(id: TId): Promise<TData>

  abstract findAndCountAll(criteria: TCriteria): Promise<PageData<TData>>

  async findAll(criteria: TCriteria): Promise<TData[]> {
    const copy = { ...criteria }

    // let's initially load just the first row in the db to see how many elements there are in total
    copy.offset = undefined
    copy.limit = undefined
    const page = await this.findAndCountAll(criteria)

    return page.rows
  }

  protected async createAuditLog(
    entity: string,
    action: string,
    record: any,
    data: any,
  ): Promise<void> {
    if (this.log) {
      let values = {}

      if (data) {
        values = {
          ...record.get({ plain: true }),
        }
      }

      await AuditLogRepository.log(
        {
          entityName: entity,
          entityId: record.id,
          action,
          values,
        },
        this.options,
      )
    }
  }

  protected async populateRelationsForRows(rows: any[]): Promise<any[]> {
    return Promise.all(rows.map((r) => this.populateRelations(r)))
  }

  protected async populateRelations(record: any): Promise<any> {
    if (!record) return record

    return record.get({ plain: true })
  }

  protected isPaginationValid(criteria: SearchCriteria): boolean {
    if (criteria.limit && criteria.offset) {
      return criteria.limit > 0 && criteria.offset >= 0
    }

    return false
  }

  protected getPaginationString(criteria: SearchCriteria): string {
    if (this.isPaginationValid(criteria)) {
      return `limit ${criteria.limit} offset ${criteria.offset}`
    }

    return ''
  }
}
